from datetime import datetime

from .definitions import DatabaseHandler
from .models import create_log_record_class


class LoggingDatabaseHandler(DatabaseHandler):
    """Custom logging handler to log standard logging records to a database."""

    @classmethod
    async def create(
        cls,
        **kwargs,
    ):
        return await cls._create(table_creator=create_log_record_class, **kwargs)

    async def get_object(self, record) -> dict:
        return dict(
            created_at=datetime.fromtimestamp(record.created),
            created_relative=record.relativeCreated,
            level=record.levelno,
            line_number=record.lineno,
            process=record.process,
            process_name=DatabaseHandler.cut(record.processName, 100),
            thread=record.thread,
            thread_name=DatabaseHandler.cut(record.threadName, 100),
            task_name=DatabaseHandler.cut(record.taskName, 100),
            module=DatabaseHandler.cut(record.module, 100),
            path_name=DatabaseHandler.cut(record.pathname, 300),
            message=DatabaseHandler.cut(record.getMessage(), 300),
            func_name=DatabaseHandler.cut(record.funcName, 100),
            stack_info=record.stack_info,
            exc_text=DatabaseHandler.cut(record.exc_text, 300),
        )



"""
Module for logging fastapi requests and responses.

Inspiration from https://github.com/azhig/fastapi-logging

Modified in many places and kept as minimal.
Included logging to database and excluded extensive console log support.
"""

from .utils import register_middleware  # noqa F401
from .utils import setup_logger  # noqa F401



"""Custom logging utilities."""

import asyncio
import logging
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import List

from medwave_extended_sql import AsyncDBManager

from .definitions import DatabaseHandler
from .formatters import ConsoleFormatter
from .handlers import LoggingDatabaseHandler

_logger = logging.getLogger("extended_logging")


class _SpecialHandler(logging.Handler):
    def __init__(self, log_event) -> None:
        super().__init__()
        self.log_event = log_event
        self.loop = asyncio.get_event_loop()

    def emit(self, record) -> None:
        asyncio.run_coroutine_threadsafe(self._alive(), self.loop)

    async def _alive(self):
        self.log_event.set()


async def _logger_monitor(
    logger: logging.Logger, name: str, handler: logging.Handler, timeout: int = 5
):
    """
    Monitor the logger for a specified timeout period.
    """
    log_event = asyncio.Event()
    handle = _SpecialHandler(log_event)
    logger.addHandler(handle)

    try:
        while True:
            try:
                await asyncio.wait_for(log_event.wait(), timeout=timeout)
                log_event.clear()
            except asyncio.TimeoutError:
                if isinstance(handler, DatabaseHandler):
                    if not handler._queue.empty():
                        log_event.set()
                        _logger.warning(
                            f"No logs appeared within the timeout period ({timeout}). "
                            f"Yet it has a DataBase handler with queue of size {handler._queue.qsize()}."
                        )

                    else:
                        _logger.warning(
                            f"No logs appeared within the timeout period ({timeout}). Stopping logger."
                        )
                        await handler.shutdown()
                        _logger.warning(f"Logger {name} - stopped")
                        break
                else:
                    _logger.warning(
                        f"No logs appeared within the timeout period ({timeout}). Stopping logger."
                    )
                    break

    finally:
        logger.removeHandler(handle)


async def _init_logger(
    name: str,
    original_name: str,
    level: str = "INFO",
    formatter: logging.Formatter = None,
    handler: logging.Handler = logging.StreamHandler(),
    timeout: int = 5,
) -> None:
    """
    Set up and return a custom logger with the specified name and level.

    :param name: The name of the logger to be logged to as started.
    :type name: str
    :param original_name: The name of the logger.
    :type original_name: str
    :param level: The logging level.
    :type level: str
    :param formatter: The logging formatter.
    :type formatter: logging.Formatter
    :param handler: The logging handler.
    :type handler: logging.Handler
    :param timeout: The timeout for the queue.
    :type timeout: int
    """
    logger = logging.getLogger(original_name)
    logger.setLevel(level)

    if formatter is not None:
        handler.setFormatter(formatter)

    que = Queue()
    logger.addHandler(QueueHandler(que))
    listener = QueueListener(que, handler)

    try:
        listener.start()
        _logger.debug(f"Logger {name} has started ({level})")
        await _logger_monitor(
            logger=logger, name=name, timeout=timeout, handler=handler
        )
    except Exception as e:
        _logger.error(f"Logger {name} failed: {e}")
    finally:
        _logger.debug(f"Logger {name} is shutting down")
        listener.stop()


async def _safely_start_logger(name: str, **kwargs):
    """Safely start a logger."""
    _logger.debug(f"Attempting starting logger {name}")
    logger_task = asyncio.create_task(
        _init_logger(name=name, **kwargs), name=f"{name}_logger"
    )
    await asyncio.sleep(0)
    return logger_task


async def setup_logger(
    name: str = None,
    level: str = "INFO",
    console_formatter: logging.Formatter = None,
    db_manager: AsyncDBManager = None,
    schema: str = "logs",
    console_: bool = True,
    database_: bool = True,
    drop_: bool = False,
    timeout: int = 99999,
    db_workers: int = 8,
) -> List[asyncio.Task]:
    """
    Set up and return a custom logger with the specified name and level.

    :param name: Name of the logger.
    :type name: str
    :param level: Logging level, e.g., 'INFO', 'DEBUG'.
    :type level: str
    :param console_formatter: Formatter instance to use for loging messages into console.
    :type console_formatter: logging.Formatter
    :param db_manager: AsyncDBManager to use for logging messages into database.
    :type db_manager: AsyncDBManager
    :param schema: The schema to use for the database table.
    :type schema: str
    :param console_: Whether to log to console (add console handler)
    :type console_: bool
    :param database_: Whether to log to database (add database handler)
    :type database_: bool
    :param drop_: Whether to drop the table before creating it.
    :type drop_: bool
    :param timeout: The timeout for the queue.
    :type timeout: int
    :param db_workers: The number of workers to use for the database handler.
    :type db_workers: int
    :return: Configured logger instance.
    :rtype: logging.Logger
    """
    if not console_ and not database_:
        raise ValueError("At least one of console_ or database_ must be True.")
    tasks = []

    # add console handler
    if console_:
        if console_formatter is None:
            console_formatter = ConsoleFormatter(name=name)

        console_handler = logging.StreamHandler()
        tasks.append(
            await _safely_start_logger(
                name=name + "_console",
                original_name=name,
                level=level,
                formatter=console_formatter,
                handler=console_handler,
                timeout=timeout,
            )
        )

    # add database handler
    if database_:
        db_handler = await LoggingDatabaseHandler.create(
            db_manager=db_manager,
            table_name=name,
            drop_=drop_,
            schema=schema,
            timeout=timeout,
            num_workers=db_workers,
        )

        tasks.append(
            await _safely_start_logger(
                name=name + "_database",
                original_name=name,
                level=level,
                handler=db_handler,
                timeout=timeout,
            )
        )

    return tasks



"""Custom logging formatters"""

import logging


class ConsoleFormatter(logging.Formatter):
    """Logging Formatter to add colors to the levelname and include function name"""

    grey = "\x1b[38;21m"
    blue = "\x1b[34m"
    light_blue = "\x1b[94m"
    green = "\x1b[32m"
    yellow = "\x1b[33m"
    red = "\x1b[31m"
    purple = "\x1b[35m"
    reset = "\x1b[0m"

    LEVEL_COLORS = {
        logging.DEBUG: blue,
        logging.INFO: green,
        logging.WARNING: yellow,
        logging.ERROR: red,
        logging.CRITICAL: purple,
    }

    def __init__(
        self,
        style: str = "%(asctime)s - %(levelname)s - (%(name)s, %(taskName)s) - %(message)s",
        datefmt: str = "%Y-%m-%d %H:%M:%S",
        name: str = "",
    ):
        super().__init__()
        self._style._fmt = style
        self.datefmt = datefmt

    def format(self, record):
        """Format the log record with colors and function name"""

        # Determine the color based on the log level
        color_code = ConsoleFormatter.LEVEL_COLORS.get(
            record.levelno, ConsoleFormatter.grey
        )

        # Apply color and reset codes, along with the log level name
        colored_level_name = (
            f"{color_code}{record.levelname}{ConsoleFormatter.reset:<17s}"
        )

        # Replace the original log level name with the colored one
        record.levelname = colored_level_name

        # Format the rest of the log record as usual
        return super().format(record)



from datetime import datetime
from typing import Any, Dict, Optional

from sqlalchemy import JSON, REAL, BigInteger, Column, Integer, SmallInteger
from sqlmodel import Field

from medwave_extended_sql import register_class
from medwave_extended_sql.models import ExtendedSQLModel, IdMixin


class RequestLogRecordBase:
    """
    Reference to class defined by :func:`create_request_log_record_class`.
    Just for type hinting.
    """


def create_request_log_record_class(
    table_name: str, schema: str = "logs"
) -> RequestLogRecordBase:
    """
    Creates a RequestLogRecordBase class with the given table name and schema.

    :param table_name: The name of the table.
    :type table_name: str
    :param schema: The name of the schema.
    :type schema: str
    :return: The RequestLogRecordBase class that has table representation.
    :rtype: RequestLogRecordBase
    """
    attrs = {
        "created_at": Field(
            default_factory=datetime.utcnow, nullable=False, index=True
        ),
        "created_relative": Field(sa_column=Column(REAL, nullable=False)),
        "level": Field(sa_column=Column(SmallInteger, nullable=False, index=True)),
        "line_number": Field(sa_column=Column(SmallInteger, nullable=False)),
        "process": Field(sa_column=Column(Integer, nullable=False, index=True)),
        "process_name": Field(default=None, max_length=100),
        "thread": Field(sa_column=Column(BigInteger, nullable=False)),
        "thread_name": Field(default=None, max_length=100),
        "task_name": Field(default=None, max_length=100),
        "request_uri": Field(nullable=False, max_length=350),
        "request_protocol": Field(nullable=False, max_length=50),
        "request_method": Field(nullable=False, max_length=10),
        "request_path": Field(nullable=False, max_length=300),
        "request_host": Field(nullable=False, max_length=50),
        "request_size": Field(sa_column=Column(Integer, nullable=False)),
        "request_content_type": Field(nullable=False, max_length=100),
        "request_headers": Field(sa_column=Column(JSON, nullable=False)),
        "request_body": Field(sa_column=Column(JSON, nullable=False)),
        "remote_host": Field(nullable=False, max_length=100),
        "response_status_code": Field(sa_column=Column(SmallInteger, nullable=False)),
        "response_size": Field(sa_column=Column(Integer, nullable=False)),
        "response_headers": Field(sa_column=Column(JSON, nullable=False)),
        "response_body": Field(sa_column=Column(JSON, nullable=False)),
        "duration": Field(sa_column=Column(SmallInteger, nullable=False)),
        "__annotations__": {
            "created_at": datetime,
            "created_relative": float,
            "level": int,
            "line_number": int,
            "process": int,
            "process_name": Optional[str],
            "thread": int,
            "thread_name": Optional[str],
            "task_name": Optional[str],
            "request_uri": str,
            "request_protocol": str,
            "request_method": str,
            "request_path": str,
            "request_host": str,
            "request_size": int,
            "request_content_type": str,
            "request_headers": Dict[str, Any],
            "request_body": Dict[str, Any],
            "remote_host": str,
            "response_status_code": int,
            "response_size": int,
            "response_headers": Dict[str, Any],
            "response_body": Dict[str, Any],
            "duration": int,
        },
    }
    return register_class(
        attrs=attrs,
        table_name=table_name,
        schema=schema,
        bases=(
            IdMixin,
            ExtendedSQLModel,
        ),
    )



from datetime import datetime
from typing import Optional

from sqlalchemy import REAL, BigInteger, Column, Integer, SmallInteger, UnicodeText
from sqlmodel import Field

from medwave_extended_sql import register_class
from medwave_extended_sql.models import ExtendedSQLModel, IdMixin


class LogRecordBase:
    """
    Reference to class defined by :func:`create_log_record_class`.
    Just for type hinting.
    """


def create_log_record_class(table_name: str, schema: str = "logs") -> LogRecordBase:
    """
    Creates a LogRecord class with the given table name and schema.

    :param table_name: The name of the table.
    :type table_name: str
    :param schema: The name of the schema.
    :type schema: str
    :return: The LogRecord class that has table representation.
    """
    attrs = {
        "created_at": Field(
            default_factory=datetime.utcnow, nullable=False, index=True
        ),
        "created_relative": Field(sa_column=Column(REAL, nullable=False)),
        "level": Field(sa_column=Column(SmallInteger, nullable=False, index=True)),
        "line_number": Field(sa_column=Column(SmallInteger, nullable=False)),
        "process": Field(sa_column=Column(Integer, nullable=False, index=True)),
        "process_name": Field(default=None, max_length=100),
        "thread": Field(sa_column=Column(BigInteger, nullable=False)),
        "thread_name": Field(default=None, max_length=100),
        "task_name": Field(default=None, max_length=100),
        "module": Field(nullable=False, max_length=100),
        "path_name": Field(default=None, max_length=300),
        "message": Field(nullable=False, max_length=300),
        "func_name": Field(nullable=False, max_length=100),
        "stack_info": Field(default=None, sa_column=Column(UnicodeText)),
        "exc_text": Field(default=None, max_length=300),
        "__annotations__": {
            "created_at": datetime,
            "created_relative": float,
            "level": int,
            "line_number": int,
            "process": int,
            "process_name": Optional[str],
            "thread": int,
            "thread_name": Optional[str],
            "task_name": Optional[str],
            "module": str,
            "path_name": Optional[str],
            "message": str,
            "func_name": str,
            "stack_info": Optional[str],
            "exc_text": Optional[str],
        },
    }

    return register_class(
        attrs=attrs,
        table_name=table_name,
        schema=schema,
        bases=(
            IdMixin,
            ExtendedSQLModel,
        ),
    )



"""Cusom logging utilities."""

import asyncio
import logging
from typing import List

from fastapi import FastAPI

from medwave_extended_sql import AsyncDBManager

from ..formatters import ConsoleFormatter
from ..utils import _safely_start_logger
from .handlers import TrafficDatabaseHandler
from .middlewares import LoggingMiddleware


async def setup_logger(
    name: str = None,
    level: str = "INFO",
    console_formatter: logging.Formatter = None,
    db_manager: AsyncDBManager = None,
    schema: str = "logs",
    console_: bool = True,
    database_: bool = True,
    drop_: bool = False,
    timeout: int = 99999,
    db_workers: int = 8,
) -> List[asyncio.Task]:
    """
    Set up and return a custom logger with the specified name and level.

    :param name: Name of the logger.
    :type name: str
    :param level: Logging level, e.g., 'INFO', 'DEBUG'.
    :type level: str
    :param console_formatter: Formatter instance to use for loging messages into console.
    :type console_formatter: logging.Formatter
    :param db_manager: AsyncDBManager to use for logging messages into database.
    :type db_manager: AsyncDBManager
    :param schema: The schema to use for the database table.
    :type schema: str
    :param console_: Whether to log to console (add console handler)
    :type console_: bool
    :param database_: Whether to log to database (add database handler)
    :type database_: bool
    :param drop_: Whether to drop the table before creating it.
    :type drop_: bool
    :param timeout: The timeout for the queue.
    :type timeout: int
    :param db_workers: The number of workers to use for the database handler.
    :type db_workers: int
    :return: Configured logger instance.
    :rtype: logging.Logger
    """
    if not console_ and not database_:
        raise ValueError("At least one of console_ or database_ must be True.")
    tasks = []

    # add console handler
    if console_:
        if console_formatter is None:
            console_formatter = ConsoleFormatter(name=name)

        console_handler = logging.StreamHandler()
        tasks.append(
            await _safely_start_logger(
                name=name + "_console",
                original_name=name,
                level=level,
                formatter=console_formatter,
                handler=console_handler,
                timeout=timeout,
            )
        )

    # add database handler
    if database_:
        db_handler = await TrafficDatabaseHandler.create(
            db_manager=db_manager,
            table_name=name,
            drop_=drop_,
            schema=schema,
            timeout=timeout,
            num_workers=db_workers,
        )

        tasks.append(
            await _safely_start_logger(
                name=name + "_database",
                original_name=name,
                level=level,
                handler=db_handler,
                timeout=timeout,
            )
        )

    return tasks


def register_middleware(app: FastAPI, logger_name: str) -> None:
    """
    Register the logging middleware to the FastAPI app.

    :param app: The FastAPI app instance.
    :type app: FastAPI
    :param logger_name: The name of the logger.
    :type logger_name: str
    """
    app.middleware("http")(
        LoggingMiddleware(
            logger_name=logger_name,
        )
    )



import http
import json
import logging
import math
import time
from logging import ERROR, INFO, WARNING
from typing import Any, Dict, Optional

from fastapi import Request, Response
from starlette.middleware.base import RequestResponseEndpoint
from starlette.types import Message

EMPTY_VALUE = ""


class Middleware:
    def __init__(self) -> None:
        Middleware.reset_logging_handlers()

    async def _log(
        self,
        level: int,
        message: str,
        extra_fields: Dict[str, Any],
        exception_object: Optional[Exception],
    ):
        raise NotImplementedError()

    async def __call__(
        self, request: Request, call_next: RequestResponseEndpoint, *args, **kwargs
    ):
        start_time = time.time()
        exception_object = None

        # Request Side
        try:
            raw_request_body = await request.body()
            await Middleware.set_body(request, raw_request_body)
            raw_request_body = await Middleware.get_body(request)
            request_body = raw_request_body.decode()
        except Exception:
            request_body = EMPTY_VALUE

        server: tuple = request.get("server", ("localhost", 9090))
        request_headers: dict = dict(request.headers.items())

        # Response Side
        try:
            response = await call_next(request)
        except Exception as ex:
            response_body = bytes(http.HTTPStatus.INTERNAL_SERVER_ERROR.phrase.encode())
            response = Response(
                content=response_body,
                status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR.real,
            )
            exception_object = ex
            response_headers = {}
        else:
            response_headers = dict(response.headers.items())
            response_body = b""
            async for chunk in response.body_iterator:  # type: ignore
                response_body += chunk
            response = Response(
                content=response_body,
                status_code=response.status_code,
                headers=dict(response.headers),
                media_type=response.media_type,
            )

        duration: int = math.ceil((time.time() - start_time) * 1000)

        # Initialization and formation of fields for request-response
        extra_fields = dict(
            request_uri=str(request.url),
            request_protocol=await Middleware.get_protocol(request),
            request_method=request.method,
            request_path=request.url.path,
            request_host=f"{server[0]}:{server[1]}",
            request_size=int(request_headers.get("content-length", 0)),
            request_content_type=request_headers.get("content-type", EMPTY_VALUE),
            request_headers=json.dumps(request_headers),
            request_body=request_body,
            remote_host=f"{request.client[0]}:{request.client[1]}",
            response_status_code=response.status_code,
            response_size=int(response_headers.get("content-length", 0)),
            response_headers=json.dumps(response_headers),
            response_body=response_body.decode(),
            duration=duration,
        )

        if exception_object:
            level = ERROR
        elif 0 <= response.status_code <= 299:
            level = INFO
        elif 300 <= response.status_code <= 399:
            level = WARNING
        else:
            level = ERROR

        message = (
            f"{'Error' if level == ERROR else 'Response'} "
            f"with code {response.status_code} "
            f"for request {request.method} {str(request.url)}"
        )

        await self._log(
            level=level,
            message=message,
            extra_fields=extra_fields,
            exception_object=exception_object,
        )
        return response

    @staticmethod
    async def get_protocol(request: Request) -> str:
        protocol = str(request.scope.get("type", ""))
        http_version = str(request.scope.get("http_version", ""))
        if protocol.lower() == "http" and http_version:
            return f"{protocol.upper()}/{http_version}"
        return EMPTY_VALUE

    @staticmethod
    async def get_body(request: Request) -> bytes:
        body = await request.body()
        await Middleware.set_body(request, body)
        return body

    @staticmethod
    async def set_body(request: Request, body: bytes) -> None:
        async def receive() -> Message:
            return {"type": "http.request", "body": body}

        request._receive = receive

    @staticmethod
    def reset_logging_handlers() -> None:
        """Remove all default fastapi and gunicorn loggers."""
        loggers = (
            logging.getLogger(name)
            for name in logging.root.manager.loggerDict
            if name.startswith(("uvicorn.", "gunicorn."))
        )
        for logger in loggers:
            logger.handlers = []


class LoggingMiddleware(Middleware):
    """Logging Malware with support for logging module"""

    def __init__(self, logger_name: str, reload_logger_: bool = True) -> None:
        super().__init__()

        self.logger_name = logger_name
        self.logger = logging.getLogger(logger_name)
        self.reload_logger_ = reload_logger_

    def _get_logger(self):
        if self.reload_logger_:
            return logging.getLogger(self.logger_name)
        return self.logger

    async def _log(
        self,
        level: int,
        message: str,
        extra_fields: Dict[str, Any],
        exception_object: Exception,
    ):
        self._get_logger().log(
            level,
            message,
            extra=extra_fields,
            exc_info=exception_object,
        )



"""Custom logging handlers for logging fastapi traffic."""

from datetime import datetime

from ..definitions import DatabaseHandler
from .models import create_request_log_record_class


class TrafficDatabaseHandler(DatabaseHandler):
    """Custom logging handler to log traffic into database."""

    @classmethod
    async def create(
        cls,
        **kwargs,
    ):
        return await cls._create(
            table_creator=create_request_log_record_class, **kwargs
        )

    async def get_object(self, record) -> dict:
        return dict(
            created_at=datetime.fromtimestamp(record.created),
            created_relative=record.relativeCreated,
            level=record.levelno,
            line_number=record.lineno,
            process=record.process,
            process_name=record.processName[:100],
            thread=record.thread,
            thread_name=record.threadName[:100],
            task_name=record.taskName[:100],
            request_uri=record.request_uri[:350],
            request_protocol=record.request_protocol[:50],
            request_method=record.request_method[:10],
            request_path=record.request_path[:300],
            request_host=record.request_host[:50],
            request_size=record.request_size,
            request_content_type=record.request_content_type[:100],
            request_headers=record.request_headers,
            request_body=record.request_body,
            remote_host=record.remote_host[:100],
            response_status_code=record.response_status_code,
            response_size=record.response_size,
            response_headers=record.response_headers,
            response_body=record.response_body,
            duration=record.duration,
        )



"""Custom logging handlers."""

import asyncio
import inspect
import logging
import os

from medwave_extended_sql import AsyncDBManager


class DatabaseHandler(logging.Handler):
    """Custom logging handler to save records to a database."""

    SIG_KILL = "KILLME"

    def __init__(
        self,
        db_manager: AsyncDBManager,
        table_name: str,
        timeout: int = 5,
        num_workers: int = 8,
    ):
        super().__init__()

        self.db_manager = db_manager
        self.loop = asyncio.get_event_loop()
        self.name = f"{table_name}_db_handler_queue"
        self.timeout = timeout
        self.num_workers = num_workers

        self._queue = asyncio.Queue()
        self._logger = logging.getLogger("extended_logging")

    @classmethod
    async def _create(
        cls,
        db_manager: AsyncDBManager,
        table_creator: callable,
        table_name: str,
        schema: str = "logs",
        drop_: bool = False,
        timeout: int = 5,
        num_workers: int = 8,
    ):
        """
        Create a new instance of the DatabaseHandler.

        :param db_manager: The database manager.
        :type db_manager: AsyncDBManager
        :param table_creator: The function to create the table.
        Will be fed with `table_name` and `schema`.
        :type table_creator: callable
        :param table_name: The name of the table to log to.
        :type table_name: str
        :param schema: The schema to use for the database table.
        :type schema: str
        :param drop_: Whether to drop the table.
        :type drop_: bool
        :param timeout: The timeout for the queue.
        :type timeout: int
        :param num_workers: The number of workers to use.
        :type num_workers: int
        :return: The new instance of the DatabaseHandler.
        :rtype: DatabaseHandler
        """
        self = cls(
            db_manager=db_manager,
            table_name=table_name,
            timeout=timeout,
            num_workers=num_workers,
        )

        self._logger.debug(f"{self.name}: initializing table and schemas")
        self._table_class = table_creator(table_name=table_name, schema=schema)
        try:
            await self.db_manager.register_schema(schema=schema)
            await self.db_manager.initialize(drop_=drop_, model=self._table_class)
            await asyncio.sleep(0)
        except Exception as e:
            self._logger.error(f"error initializing table: {e}")

        self._logger.debug(f"{self.name}: initializing queue listener")
        self._tasks = []
        for i in range(num_workers):
            worker_name = f"{self.name}_worker_{i+1}"
            self._tasks.append(
                asyncio.create_task(self._process_queue(i + 1), name=worker_name)
            )

        return self

    async def _process_queue(self, worker_name: str):
        while True:
            record = await self._queue.get()
            self._logger.debug(f"got: {record}")

            if record == DatabaseHandler.SIG_KILL:
                self._logger.debug("received SIG_KILL")
                self._queue.task_done()
                break

            await self._write_to_db(record)
            self._queue.task_done()

    def emit(self, record):
        if record.pathname.strip().endswith(
            "/src/utils/logging/fastapi/middlewares.py"
        ):
            return

        func_name = DatabaseHandler.get_func_name()
        if func_name is not None:
            record.funcName = func_name

        try:
            t = asyncio.run_coroutine_threadsafe(self._queue.put(record), self.loop)
            t.result()
        except Exception as e:
            self._logger.error(f"error putting record to queue: {e}")

        self._logger.debug(f"{self.name}: queue size: {self._queue.qsize()}")

    async def get_object(self, record) -> dict:
        """
        Called before writing to database, should return dict that will be passed to
        `table_class` constructor retrieved from `table_creator` callable.
        """
        raise NotImplementedError

    async def _write_to_db(self, record):
        try:
            await self.db_manager.insert(
                o=self._table_class(**await self.get_object(record=record))
            )
            self._logger.debug(f"saved: {record}")
        except Exception as e:
            self._logger.error(f"error saving record: {e}")

    async def shutdown(self):
        self._logger.debug(f"{self.name}: Attempting shutdown.")

        for _ in range(self.num_workers):
            await self._queue.put(DatabaseHandler.SIG_KILL)
        await self._queue.join()

        for task in self._tasks:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                self._logger.debug(f"{self.name}: task {task.get_name()} cancelled.")

    @staticmethod
    def get_func_name() -> str:
        """
        Get the name of the function that called the logger.

        :return: The name of the function.
        :rtype: str
        """
        stack = inspect.stack()
        for frame in stack:
            if (
                not frame.function.startswith("_")
                and frame.function != "get_func_name"
                and frame.function not in ["emit", "format"]
                and "logging" not in frame.filename.split(os.sep)
            ):
                return frame.function

    @staticmethod
    def cut(o: str, length: int = 300) -> str:
        """
        Cut the string to the specified length if it is not None.

        :param o: The string to cut.
        :type o: str
        :param length: The length to cut to.
        :type length: int
        :return: The cut string.
        :rtype: str
        """
        return o[:length] if o is not None else None
